local famBags = LibStub('AceAddon-3.0'):NewAddon('famBags', 'AceEvent-3.0', 'AceConsole-3.0', 'AceHook-3.0')
local L = LibStub('AceLocale-3.0'):GetLocale('famBags')
local LSM = LibStub('LibSharedMedia-3.0')
local lbf = LibStub("LibButtonFacade", true)
famBags.L = L
famBags.SortedCategories = {}
famBags.BagIndexes = {}
famBags.Inventory = {}
famBags.CHANGING_PROFILE = nil
famBags.BAGS_NEED_LAYOUT = nil
famBags.BANK_NEED_LAYOUT = nil
famBags.EMPTIES_NEED_LAYOUT = {
   bags = true,
   bank = true,
}
famBags.BagTypes = {
   [0] = 'any',
   [1] = 'quiver',
   [2] = 'ammo',
   [4] = 'soul',
   [8] = 'leather',
   [16] = 'inscription',
   [32] = 'herbs',
   [64] = 'enchanting',
   [128] = 'engineering',
   [256] = 'key',
   [512] = 'gem',
   [1024] = 'mining',
}
local db

-- :::::::::::: famBags.Categories ::::::::::::: -------------------------------------------------------------------------

famBags.Types = {
   [L['Armor']] = { L['Miscellaneous'], L['Cloth'], L['Leather'], L['Mail'], L['Plate'], L['Shields'], L['Librams'], L['Idols'], L['Totems'], L['Sigils'] },
   [L['Consumable']] = { L['Food & Drink'], L['Potion'], L['Elixir'], L['Flask'], L['Bandage'], L['Item Enhancement'], L['Scroll'], L['Other'], L['Consumable'] },
   [L['Container']] = { L['Bag'], L['Enchanting Bag'], L['Engineering Bag'], L['Gem Bag'], L['Herb Bag'], L['Mining Bag'], L['Soul Bag'], L['Leatherworking Bag'] },
   [L['Gem']] = { L['Blue'], L['Green'], L['Orange'], L['Meta'], L['Prismatic'], L['Purple'], L['Red'], L['Simple'], L['Yellow'] },
   [L['Miscellaneous']] = { L['Junk'], L['Reagent'], L['Pet'], L['Holiday'], L['Mount'], L['Other'] },
   [L['Money']] = {},
   [L['Reagent']] = {},
   [L['Recipe']] = { L['Alchemy'], L['Blacksmithing'], L['Book'], L['Cooking'], L['Enchanting'], L['Engineering'], L['First Aid'], L['Leatherworking'], L['Tailoring'] },
   [L['Quest']] = {},
   [L['Trade Goods']] = { L['Armor Enchantment'], L['Cloth'], L['Devices'], L['Elemental'], L['Enchanting'], L['Explosives'], L['Herb'], L['Jewelcrafting'], L['Leather'],
                        L['Materials'], L['Meat'], L['Metal & Stone'], L['Other'], L['Parts'], L['Trade Goods'], L['Weapon Enchantment'] },
   [L['Weapon']] = { L['Bows'], L['Crossbows'], L['Daggers'], L['Guns'], L['Fishing Poles'], L['Fist Weapons'], L['Miscellaneous'], L['One-Handed Axes'], L['One-Handed Maces'],
                   L['One-Handed Swords'], L['Polearms'], L['Staves'], L['Thrown'], L['Two-Handed Axes'], L['Two-Handed Maces'], L['Two-Handed Swords'], L['Wands'] },
}

famBags.Criterias = {
   [L['Quality']] = { options = {L['unknown'], '0.'..ITEM_QUALITY0_DESC, '1.'..ITEM_QUALITY1_DESC, '2.'..ITEM_QUALITY2_DESC, '3.'..ITEM_QUALITY3_DESC, '4.'..ITEM_QUALITY4_DESC, '5.'..ITEM_QUALITY5_DESC, 
                     '6.'..ITEM_QUALITY6_DESC, '7.'..ITEM_QUALITY7_DESC, } },
   [L['Equipment Slot']] = { options = {INVTYPE_AMMO, INVTYPE_HEAD, INVTYPE_NECK, INVTYPE_SHOULDER, INVTYPE_BODY, INVTYPE_CHEST, INVTYPE_ROBE, INVTYPE_WAIST, INVTYPE_LEGS, INVTYPE_FEET, 
                            INVTYPE_WRIST, INVTYPE_HAND, INVTYPE_FINGER, INVTYPE_TRINKET, INVTYPE_CLOAK, INVTYPE_WEAPON, INVTYPE_SHIELD, INVTYPE_2HWEAPON, 
                            INVTYPE_WEAPONMAINHAND, INVTYPE_WEAPONOFFHAND, INVTYPE_HOLDABLE, INVTYPE_RANGED, INVTYPE_THROWN, INVTYPE_RANGEDRIGHT, INVTYPE_RELIC, INVTYPE_TABARD }, },
   [L['Bound Status']] = { options = {L['Binds when picked up'], L['Binds when equipped'], L['Soulbound'], L['Binds to account'], L['Account Bound']}, },
   [L['Equipment Set']] = { options = {}, },
   [L['Quest']] = { options = {}, },
   [L['Key']] = { options = {}, },
   [L['Empty']] = { options = {}, },
   [L['Usable']] = { options = {}, },
   [L['New']] = { options = {}, },
   [L['Incremented']] = { options = {}, },
   [L['Others']] = { options = {}, },
   [L['Name']] = { options = {L['Equal'], L['Contains'], L['Starts with'], L['Ends with']}, valueType = 'text', },
}

-- :::::::::::: DB ::::::::::::: -------------------------------------------------------------------------

function famBags:GetBGDefaults()
   return {
      type = 1, -- Solid
      color1 = {0,0,0,0.7},
      color2 = {0,0,0,0.7},
      texture = nil,
      inset = 0,
      br_texture = 'None',
      br_color = {1,1,1,1},
      br_size = 10,
   }
end

function famBags:GetTextDefaults()
   return {
      font = '',
      size = 11,
         flags = {
         ['OUTLINE'] = false,
         ['THICKOUTLINE'] = false,
         ['MONOCHROME'] = false,
         ['SHADOW'] = true,
      },
      color = {1,1,1},
      halign = 'LEFT',
      valign = 'CENTER',
      padding = 1,
   }
end

function famBags:GetTitleDefaults()
   return {
      is_inset = false,
      inset = 0,
      bg = self:GetBGDefaults(),
      text = self:GetTextDefaults()
   }
end

local defaults = {
   profile =  {
      layout = {
         skins = {
            bags = {},
         },
         general ={
            strata = 'HIGH',
            sections = 'categories',
            ordered = true,
            show_keys = true,
            full_layout = false,
            update_time = 0.7,
         },
         sections = {
            compact = true,
            hpadding = 5,
            vpadding = 5,
            hspacing = 5,
            vspacing = 5,
         },
         buttons = {
            hpadding = 5,
            vpadding = 5,
            hspacing = 5,
            vspacing = 5,
            size = 0.7,
            quality_glow = 'round1',
            quality_glow_alpha = 1,
            quality_glow_inset = 0,
            indicators_enabled = true,
            indicators_size = 1,
            indicators_offset = 2,
         },
         stacks = {
            only_stackable = false,
            only_full = false,
         },
         counts = famBags:GetTextDefaults(),
         empties = {
            bg = famBags:GetBGDefaults(),
            hpadding = 5,
            vpadding = 5,
            vpos = 'TOP',
            hpos = 'RIGHT',
            no_empty_text = 'x',
            no_empty_color = {1,0,0},
         },
      },
      mainwindow = {
         layout = {
            buttons_x_row = 10,
            halign = 'RIGHT',
            valign = 'TOP',
            hspacing = 40,
            vspacing = 40,
            fixed_width = true,
         },
         bg = famBags:GetBGDefaults(),
         title = famBags:GetTitleDefaults()
      },
      money = {
         enabled = true,
         show_silver = true,
         show_copper = true,
         bg = famBags:GetBGDefaults(),
         text = famBags:GetTextDefaults(),
      },
      bank = {
         layout = {
            buttons_x_row = 15,
            halign = 'LEFT',
            valign = 'TOP',
            hspacing = 40,
            vspacing = 40,
            fixed_width = true,
         },
         bg = famBags:GetBGDefaults(),
         title = famBags:GetTitleDefaults()
      },
      bankwindow = {
         enabled = true,
         bg = famBags:GetBGDefaults(),
      },
      bagswindow = {
         enabled = true,
         bg = famBags:GetBGDefaults(),
      },
      sections = {
         bags= {
            bg = famBags:GetBGDefaults(),
            title = famBags:GetTitleDefaults(),
         },
      },
      categories = {
      }
   },
}

-- :::::::::::: LOCALS ::::::::::::: -------------------------------------------------------------------------

local fix_hpad = {
   CENTER = 0,
   LEFT = 0,
   RIGHT = -2
}

local fix_vpad = {
   CENTER = 1,
   TOP = -1,
   BOTTOM = 2,
}

local function get_point(h, v, px, py)
   py = py or px
   local align
   local hpad = px + fix_hpad[h]
   local vpad = py + fix_vpad[v]
   if v == 'CENTER' then
      vpad = fix_vpad[v]
      v = ''
   end
   if h == 'CENTER' then
      hpad = fix_hpad[h]
      h = ''
   end
   if h..v == '' then
      align = 'CENTER'
   else
      align = v..h
   end
   if h == 'RIGHT' then hpad = -hpad end
   if v == 'TOP' then vpad = -vpad end
   return align, hpad, vpad
end

local function npairs(t)
	local a = {}
		for n in pairs(t) do table.insert(a, n) end
		table.sort(a)
		local i = 0      -- iterator variable
		local iter = function ()   -- iterator function
			i = i + 1
			if a[i] == nil then return nil
			else return a[i], t[a[i]]
			end
		end
	return iter
end

-- timer for OnUpdate execution control
local timerFrame = CreateFrame('Frame', 'famBagsTimerFrame', UIParent)
timerFrame:Hide()
local timer = 0
local ticktime = 1
timerFrame:SetScript("OnUpdate",function(self, elapsed)
   timer = timer + elapsed
   if timer >= ticktime then
      famBags:TimerTick()
      timer = 0
   end
end)

local DEFAULT_FONT = select(1, SystemFont_Small:GetFont())
local LOADING = true
local FIRST_BUILD = true
local UPDATE_COOLDOWNS = false
local BAGS_CHANGED = {}
local BAGS_CHANGED_COUNT = 0
local CATEGORIES_DROPDOWN_ITEMS = {
   [0] = {
      text = L['Move to category'],
      value = L['Move to category'],
      isTitle = true,
      notCheckable = true,
   }
}
-- :::::::::::: famBags ::::::::::::: -------------------------------------------------------------------------

function famBags:BuildSortedCategories()
   self.SortedCategories = {}
   for n in npairs(db.categories) do
      table.insert(self.SortedCategories, n)
   end
end

function famBags:BuildBagIndexes()
   famBags.BagIndexes = {
      bags = {},
      bank = {}
   }
   for i = 0, NUM_BAG_SLOTS do
      table.insert(famBags.BagIndexes.bags, i)
   end
   if db.layout.general.show_keys then
      table.insert(famBags.BagIndexes.bags, KEYRING_CONTAINER)
   end
   table.insert(famBags.BagIndexes.bank, BANK_CONTAINER)
   for i = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do
      table.insert(famBags.BagIndexes.bank, i)
   end
end

function famBags:BuildInventory()
   wipe(self.Inventory)
   local id, count
   for _, bag in ipairs(famBags.BagIndexes.bags) do
      local slots = GetContainerNumSlots(bag)
      if slots then
         for slot = 1, slots do
            id = GetContainerItemID(bag, slot)
            if id then
               _, count = GetContainerItemInfo(bag, slot)
               if self.Inventory[id] then
                  self.Inventory[id] = self.Inventory[id] + (count or 0)
               else
                  self.Inventory[id] = count
               end
            end
         end
      end
   end
end

function famBags:GetInventoryForItem(itemID)
   if not itemID then return 0 end
   local id, count, b
   local total = 0
   for _, bag in ipairs(famBags.BagIndexes.bags) do
      local slots = GetContainerNumSlots(bag)
      if slots then
         for slot = 1, slots do
            id = GetContainerItemID(bag, slot)
            b = nil
            if id == itemID then
               if self.Buttons[bag] then
                  b = self.Buttons[bag][slot] 
                  if b then
                     if b.info.new then
                        return 'new'
                     elseif b.info.added then
                        return 'added'
                     end
                  end
               end
               _, count = GetContainerItemInfo(bag, slot)
               total = total + (count or 0)
            end
         end
      end
   end
   return total
end

function famBags:GetDBDefaults()
   return defaults
end

function famBags:OnInitialize()
   self.db = LibStub('AceDB-3.0'):New('famBags', defaults, 'Default')
   db = self.db.profile
   self.isStackMode = true
   self.MoveToCategoryDD = CreateFrame("Frame", "famBagsMoveToCategoryDD", UIParent, "UIDropDownMenuTemplate");
   UIDropDownMenu_Initialize(self.MoveToCategoryDD, famBags.InitMoveToCategoryMenu, "MENU");
   self:RegisterChatCommand('fam', 'ChatCommand')
   self:RegisterChatCommand('fambags', 'ChatCommand')
   self:RawHook("OpenAllBags", "ToggleBags", true)
   self:RawHook("OpenBackpack", "ShowBags", true)
   --self:RawHook("CloseBackpack", "HideBags", true)
   self:RawHook("CloseSpecialWindows", "HideBagsOnEsc", true)
   self:RawHook("ContainerFrameItemButton_OnEnter", "ContainerFrameItemButtonOnEnter", true)
   self:RawHook("ContainerFrameItemButton_OnModifiedClick", "ContainerFrameItemButtonOnModifiedClick", true)
   self:RawHook("SetCurrencyBackpack", "SetCurrencyBackpack", true)
   self:RawHookScript(BankFrame, "OnEvent", "BankFrameOnEvent")
end

function famBags:OnEnable()
   self:RegisterEvent('PLAYER_ENTERING_WORLD')
   self:RegisterEvent('ITEM_LOCK_CHANGED')
   self:RegisterEvent('BAG_UPDATE_COOLDOWN')
   self:RegisterEvent('EQUIPMENT_SETS_CHANGED')
   self:RegisterEvent("BANKFRAME_OPENED")
   self:RegisterEvent("BANKFRAME_CLOSED")
   self:RegisterEvent("PLAYER_MONEY")
   self:RegisterEvent('AUCTION_HOUSE_SHOW')
   self:RegisterEvent('AUCTION_HOUSE_CLOSED')
   self:RegisterEvent("CURRENCY_DISPLAY_UPDATE");
   self:RegisterEvent("KNOWN_CURRENCY_TYPES_UPDATE");
end

function famBags:ChatCommand(input)
   if not input or input:trim() == '' then
      self:LoadOptions()
   end
end

function famBags:LoadOptions()
   if LoadAddOn('famBagsOptions') then
      InterfaceOptionsFrame_OpenToCategory('famBags')
   end
end

function famBags:FullRefresh()
   local function wipe_section(sections, wipeit)
      if not sections then return end
      for _, s in pairs(sections) do
         s:Hide()
      end
      if wipeit then
         wipe(sections)
      end
   end
   
   CloseBankFrame()
   CloseBackpack()
   db = self.db.profile
   for _, indexes in pairs(famBags.BagIndexes) do
      for _, i in ipairs(indexes) do
         if self.Buttons[i] then
            for _, b in ipairs(self.Buttons[i]) do
               b:Hide()
            end
         end
      end
   end
   wipe(self.Buttons.Current)
   if self.Sections.bags then
      wipe_section(self.Sections.bags.bags)
      wipe_section(self.Sections.bags.categories, true)
   end
   if self.Sections.bank then
      wipe_section(self.Sections.bank.bags)
      wipe_section(self.Sections.bank.categories, true)
   end
   self.famBagsMoneyFrame = nil
   self.famBagsBlizzardBankFrame = nil
   self.famBagsBagSlotsFrame = nil
   self.CHANGING_PROFILE = true
   self.EMPTIES_NEED_LAYOUT['bags'] = true
   self.EMPTIES_NEED_LAYOUT['bank'] = true
   LOADING = true
   self:PLAYER_ENTERING_WORLD()
end

function famBags:BankFrameOnEvent(frame, event)
	if (not db.bankwindow.enabled) or (event ~= 'BANKFRAME_OPENED' and event ~= 'BANKFRAME_CLOSED') then
      self.hooks[BankFrame].OnEvent(frame, event)
	end
end

function famBags:ContainerFrameItemButtonOnEnter(b)
	-- Bank main container specific code
	if b:GetParent():GetID() == BANK_CONTAINER then
   	local x = b:GetRight()
   	GameTooltip:SetOwner(b, (x >= ( GetScreenWidth() / 2)) and "ANCHOR_LEFT" or "ANCHOR_RIGHT")
		GameTooltip:SetInventoryItem("player", BankButtonIDToInvSlotID(b:GetID()))
		CursorUpdate(b)
	else
      self.hooks.ContainerFrameItemButton_OnEnter(b)
	end
end

function famBags:ContainerFrameItemButtonOnModifiedClick(b, mousebutton)
	if IsControlKeyDown() and mousebutton == 'RightButton' and b.info and b.info.id then
	   self.MoveToCategoryDD.button = b
	   ToggleDropDownMenu(1, nil, self.MoveToCategoryDD, b:GetName(), b:GetWidth()/2, b:GetHeight()/2)
	else
      self.hooks.ContainerFrameItemButton_OnModifiedClick(b, mousebutton)
	end
end

function famBags:SaveSkin(index, skinID, gloss, backdrop, colors)
   db.layout.skins[index].skinID = skinID
   db.layout.skins[index].gloss = gloss
   db.layout.skins[index].backdrop = backdrop
   db.layout.skins[index].colors = colors
end

function famBags:BFSkinCallback(skinID, gloss, backdrop, group, button, colors)
	if not group then
	   self:SaveSkin('bags', skinID, gloss, backdrop, colors)
	elseif group == 'Bags' then
	   self:SaveSkin('bags', skinID, gloss, backdrop, colors)
	end
	--hack to fix texture size for keyring button
   if self.famBagsBagSlotsFrame and self.famBagsBagSlotsFrame.KeyRingTexture then
      self.famBagsBagSlotsFrame.KeyRingTexture:SetTexCoord(0, 0.5625, 0, 0.609375)
   end
end

function famBags:PLAYER_ENTERING_WORLD()
   if not LOADING then return end
   LOADING = nil
   if not self.CHANGING_PROFILE then
      self:BuildBagIndexes()
      self:EQUIPMENT_SETS_CHANGED()
      self.LDBBags:Create()
      self.LDBCurrency:Create()
   end
   if lbf then
      self.BFBags = lbf:Group("famBags", "Bags")   
      self.BFBags:Skin(db.layout.skins.bags.skinID, db.layout.skins.bags.gloss, db.layout.skins.bags.backdrop, db.layout.skins.bags.colors)
      lbf:RegisterSkinCallback("famBags", self.BFSkinCallback, self)
   end
   self:BuildSortedCategories()
   self:UpdateBags()
   self:CreateIntefaceButtons()
   self.BagsStackButton:SetChecked(true)
   self.BankStackButton:SetChecked(true)
   if db.bagswindow.enabled then
      self.BagSlotsButton:SetChecked(false)
   end
   if not self.CHANGING_PROFILE then
      self:RegisterEvent('BAG_UPDATE')
      self:RegisterEvent('PLAYERBANKSLOTS_CHANGED')
      self:RegisterEvent('PLAYERBANKBAGSLOTS_CHANGED')
   end
end

function famBags:ITEM_LOCK_CHANGED(e, bag, slot)
   if slot and not (bag == BANK_CONTAINER and slot > NUM_BANKGENERIC_SLOTS) then
      local b = self.Buttons[bag][slot]
      self.Buttons:UpdateItemLock(b, bag, slot)
      famBags.Buttons:UpdateLock(b)
   end
end

function famBags:BAG_UPDATE_COOLDOWN(e, bag)
   UPDATE_COOLDOWNS = true
   ticktime = 0.1
   self:StartTimer()
end

function famBags:PLAYERBANKSLOTS_CHANGED(e, slot)
   if slot <= NUM_BANKGENERIC_SLOTS then
      self:BAG_UPDATE(e, BANK_CONTAINER)
   else
      self:BAG_UPDATE(e, NUM_BAG_SLOTS + slot - NUM_BANKGENERIC_SLOTS)
      if db.bankwindow.enabled then
		   BankFrameItemButton_Update(self.famBagsBlizzardBankFrame.BankBags[slot - NUM_BANKGENERIC_SLOTS])
		end
   end
end

function famBags:PLAYERBANKBAGSLOTS_CHANGED()
   if db.bankwindow.enabled then
	   UpdateBagSlotStatus()
	end
   local _, full = GetNumBankSlots()
   local f = self.famBagsBlizzardBankFrame
   if full and db.bankwindow.enabled and f and f.PurchaseInfo then
      f.PurchaseInfo:ClearAllPoints()
      f.PurchaseInfo:Hide()
      f:SetHeight(f.BankBags[1]:GetHeight() + 20)
   end
end

function famBags:BAG_UPDATE(e, bag, ...)
   if bag == KEYRING_CONTAINER and not db.layout.general.show_keys then return end
   local found
   for i=1, BAGS_CHANGED_COUNT do
      if BAGS_CHANGED[i] == bag then
         found = true
         break
      end
   end
   if not (found) then
      BAGS_CHANGED_COUNT = BAGS_CHANGED_COUNT + 1
      BAGS_CHANGED[BAGS_CHANGED_COUNT] = bag
   end
   ticktime = db.layout.general.update_time
   self:StartTimer()
end

function famBags:EQUIPMENT_SETS_CHANGED()
   if self.EquipmentSets then 
      wipe(self.EquipmentSets)
   else
      self.EquipmentSets = {}
   end
   wipe(self.Criterias[L['Equipment Set']].options)
   local setscount = GetNumEquipmentSets()
   for i=1, setscount do
      local set = GetEquipmentSetInfo(i);
      local items = GetEquipmentSetItemIDs(set);
      self.EquipmentSets[set] = {}
      table.insert(self.Criterias[L['Equipment Set']].options, set)
      for _, id in pairs(items) do
         self.EquipmentSets[set][id] = true
      end
   end
   if self.famBagsFrame then
      if IsAddOnLoaded('famBagsOptions') then
         LibStub('AceConfigRegistry-3.0'):NotifyChange('famBags Categories')
      end
      self:LayoutBagsButtons()
   end
end

function famBags:BANKFRAME_OPENED()
   self.famBankFrame:Show()
   if db.bankwindow.enabled then
   	for i=1, NUM_BANKBAGSLOTS, 1 do
   		button = _G["BankFrameBag"..i];
   		BankFrameItemButton_Update(button);
   	end
   	UpdateBagSlotStatus();
   end
   self:LayoutBagsButtons('bank')
   self:ShowBags()
end

function famBags:BANKFRAME_CLOSED()
   if self.famBankFrame and self.famBankFrame:IsVisible() then
      self.famBankFrame:Hide()
      self:HideBags()
   end
end

function famBags:AUCTION_HOUSE_SHOW()
   self:ShowBags()
end

function famBags:AUCTION_HOUSE_CLOSED()
   self:HideBags()
end

function famBags:PLAYER_MONEY()
   self:UpdateCurrency()
end

function famBags:SetCurrencyBackpack(...)
   self.hooks.SetCurrencyBackpack(...)
   self:UpdateCurrency()
end

function famBags:CURRENCY_DISPLAY_UPDATE()
   self:UpdateCurrency()
end

function famBags:KNOWN_CURRENCY_TYPES_UPDATE()
   self:UpdateCurrency()
end

function famBags:UpdateCurrency()
   local money = self:GetMoney()
   if self.famBagsMoneyFrame then
      self:SetTextLayout(self.famBagsMoneyFrame, 'moneyText', db.money.text, money, true, false)
   end
   self.LDBCurrency:Update(money)
end

function famBags:StartTimer()
   timer = 0
   timerFrame:Show()
end

function famBags:StopTimer()
   timerFrame:Hide()
end

function famBags:TimerTick()
   if UPDATE_COOLDOWNS then
      self.Buttons:UpdateButtonsCooldown()
      UPDATE_COOLDOWNS = false
   end
   if BAGS_CHANGED_COUNT > 0 then
      famBags:UpdateChangedBags()
   end
   if not(UPDATE_COOLDOWNS or #BAGS_CHANGED > 0) then
      self:StopTimer()
   end
end

function famBags:ShowBags()
   if not self.famBagsFrame:IsVisible() then
      self.famBagsFrame:Show()
      self:LayoutBagsButtons('bags')
      PlaySound("igBackPackOpen")
   end
end

function famBags:HideBagsOnEsc()
   if self.famBagsFrame and self.famBagsFrame:IsVisible() then
      self:HideBags()
      CloseBankFrame()
      return self.hooks.CloseSpecialWindows() or 1
   end
   CloseBankFrame()
   return self.hooks.CloseSpecialWindows()
end

function famBags:HideBags()
   if self.famBagsFrame and self.famBagsFrame:IsVisible() then
      self.famBagsFrame:Hide()
      PlaySound("igBackPackClose")
   end
end

function famBags:ToggleBags(force)
   if (not self.famBagsFrame:IsVisible()) or force then
      self.famBagsFrame:Show()
      self:LayoutBagsButtons('bags')
      PlaySound("igBackPackOpen")
   else
      self:HideBags()
   end
end

function famBags:MoveToCategory(category, id)
   if id then
      local c = db.categories
      for _, cat in pairs(famBags.SortedCategories) do
         if c[cat].Items and c[cat].Items[id] then
            c[cat].Items[id] = nil
            break
         end
      end
      if category ~= 'clear' then
         c[category].Items = c[category].Items or {}
         c[category].Items[id] = 1
      end
      famBags:LayoutBagsButtons()
   end
end

function famBags:InitMoveToCategoryMenu()
   if not famBags.MoveToCategoryDD.button then return end
   local i = 1
   local db = db.categories
   local id = famBags.MoveToCategoryDD.button.info.id
   UIDropDownMenu_AddButton(CATEGORIES_DROPDOWN_ITEMS[0])
   for _, c in pairs(famBags.SortedCategories) do
      CATEGORIES_DROPDOWN_ITEMS[i] = CATEGORIES_DROPDOWN_ITEMS[i] or {}

      CATEGORIES_DROPDOWN_ITEMS[i].text = c
      CATEGORIES_DROPDOWN_ITEMS[i].value = c
      CATEGORIES_DROPDOWN_ITEMS[i].arg1 = c
      CATEGORIES_DROPDOWN_ITEMS[i].arg2 = id
      CATEGORIES_DROPDOWN_ITEMS[i].colorCode = nil
      CATEGORIES_DROPDOWN_ITEMS[i].notCheckable = true
      CATEGORIES_DROPDOWN_ITEMS[i].func = famBags.MoveToCategory
      
      if db[c].Items and db[c].Items[id] then
         CATEGORIES_DROPDOWN_ITEMS[i].text = c..' '..L['[clear]']
         CATEGORIES_DROPDOWN_ITEMS[i].arg1 = 'clear'
         CATEGORIES_DROPDOWN_ITEMS[i].colorCode = '|cfff4d626'
      end
      
      UIDropDownMenu_AddButton(CATEGORIES_DROPDOWN_ITEMS[i])
      i = i + 1
   end
end

function famBags:PositionBagFromDB(frame, anchor, db)
   self:PositionBag(frame, anchor, db.halign, db.valign, db.hspacing, db.vspacing)
end

function famBags:PositionBag(frame, anchor, halign, valign, x, y)
   local align, h, v = get_point(halign, valign, x, y)
   frame:ClearAllPoints()
   frame:SetPoint(align, anchor, align, h, v)
end

function famBags:UpdateBags()
   if (not self.famBagsFrame) or self.CHANGING_PROFILE then
      self.famBagsFrame = self:CreateBagFrame('famBagsFrame', L['famBags'], UIParent, 0, 300, 300, db.mainwindow, true)
      self:PositionBag(self.famBagsFrame, UIParent, db.mainwindow.layout.halign, db.mainwindow.layout.valign, db.mainwindow.layout.hspacing, db.mainwindow.layout.vspacing)
      self.famBagsFrame:Hide()
      self.famBankFrame = self:CreateBagFrame('famBankFrame', L['Bank'], UIParent, 0, 300, 300, db.bank, true)
      self:PositionBag(self.famBankFrame, UIParent, db.bank.layout.halign, db.bank.layout.valign, db.bank.layout.hspacing, db.bank.layout.vspacing)
      self.famBankFrame:Hide()
      self:UpdateMoneyFrame()
      self:UpdateBlizzardBankFrame()
      if db.bagswindow.enabled then
         self:UpdateBagSlotsFrame()
      end
   end
   self:UpdatePlayerBags()
   self.LDBBags:Update()
   self.LDBCurrency:Update()
end

function famBags:UpdateMoneyFrame()
   if db.money.enabled then
      if not self.famBagsMoneyFrame then
         self.famBagsMoneyFrame = self:CreateMoneyFrame(db.money)
      end
      self.famBagsMoneyFrame:Show()
   else
      if self.famBagsMoneyFrame then
         self.famBagsMoneyFrame:Hide()
      end
   end
end

function famBags:UpdateBlizzardBankFrame()
   if db.bankwindow.enabled then
      if not self.famBagsBlizzardBankFrame then
         self.famBagsBlizzardBankFrame = self:CreateBlizzardBankFrame(db.bankwindow)
      end
      self.famBagsBlizzardBankFrame:Show()
   elseif self.famBagsBlizzardBankFrame then
      self.famBagsBlizzardBankFrame:Hide()
   end
end

function famBags:UpdateBagSlotsFrame()
   if not self.famBagsBagSlotsFrame then
      self.famBagsBagSlotsFrame = self:CreateBagSlotsFrame(db.bagswindow)
   end
   if self.famBagsBagSlotsFrame.KeyRingTexture then
      self.famBagsBagSlotsFrame.KeyRingTexture:SetTexCoord(0, 0.5625, 0, 0.609375)
   end
   self.famBagsBagSlotsFrame:Hide()
end

function famBags:UpdateChangedBags()
   self.LDBBags:Update()
   if FIRST_BUILD or not self.famBagsFrame then
      famBags:BuildInventory()
      BAGS_CHANGED_COUNT = 0
      FIRST_BUILD = nil
      return
   end
   local i
   local updatebags
   local updatebank
   local bagchanged
   famBags.BAGS_NEED_LAYOUT = nil
   famBags.BANK_NEED_LAYOUT = nil
   for k = 1, BAGS_CHANGED_COUNT do
      i = BAGS_CHANGED[k]
      bagchanged = nil
      local slots = GetContainerNumSlots(i)
      local bagType = select(2, GetContainerNumFreeSlots(i))
      if self.Buttons and self.Buttons[i] then
         if i ~= KEYRING_CONTAINER and bagType ~= self.Buttons[i].bagType then
            bagchanged = (tContains(famBags.BagIndexes.bags, i)) and 'bags' or 'bank'
            self.Buttons[i].bagType = bagType
            famBags.EMPTIES_NEED_LAYOUT[bagchanged] = true
         end
         if #self.Buttons[i] > slots then
            for j = slots + 1, #self.Buttons[i] do
               self.Buttons[i][j]:Hide()
            end
         end
      end
      self:UpdateBagButtons(i, slots, bagchanged)
      if tContains(famBags.BagIndexes.bags, i) then
         updatebags = true
      elseif tContains(famBags.BagIndexes.bank, i) then
         updatebank = true
      end
   end
   BAGS_CHANGED_COUNT = 0
   if updatebags then
      if db.layout.general.full_layout or famBags.BAGS_NEED_LAYOUT then
         famBags:LayoutBagsButtons('bags')
         famBags:BuildInventory()
      else
         famBags:LightLayoutBagsButtons('bags')
      end
   end
   if updatebank then
      if db.layout.general.full_layout or famBags.BANK_NEED_LAYOUT then
         famBags:LayoutBagsButtons('bank')
      else
         famBags:LightLayoutBagsButtons('bank')
      end
   end
end

function famBags:UpdatePlayerBags()
   self.Sections:CreateEmpty('bags')
   self.Sections:CreateEmpty('bank')
   self.Sections:BuildBagsSections('bags')
   self.Sections:BuildBagsSections('bank')
   if self.CHANGING_PROFILE then
      self.BAGS_NEED_LAYOUT = true
      self:UpdateAllButtons()
      famBags:LayoutBagsButtons()
   end
end

function famBags:UpdateBagButtons(bag, slots, forceIndicators)
   slots = slots or GetContainerNumSlots(bag)
   if slots then
      for i=1, slots do
         self.Buttons:Create(bag, i, forceIndicators)
      end
   end
end

function famBags:UpdateEmptiesButtons(bag)
   if not self.EMPTIES_NEED_LAYOUT[bag] then return end
   if db.layout.general.sections ~= 'bags' or bag == 'bags' then
      famBags.Buttons:CreateEmpty(0, bag, 0)
   else
      famBags.Buttons:CreateEmpty(BANK_CONTAINER, bag, 0)
   end
   for n in pairs(famBags.Buttons.empties[bag]) do
      if n ~= 0 or (db.layout.general.sections == 'bags' and bag == 'bank') then
         famBags.Buttons.empties[bag][n].info.needUpdate = true
      end
   end
   local n, bagType
   for _, i in ipairs(self.BagIndexes[bag]) do
      n, bagType = GetContainerNumFreeSlots(i)
      if i == KEYRING_CONTAINER then
         bagType = 256
      elseif not (bagType and famBags.BagTypes[bagType]) then
         bagType = 0
      end
      if db.layout.general.sections ~= 'bags' then
         if bagType ~= 0 then
            local b = famBags.Buttons:CreateEmpty(bagType, bag, bagType)
            b.info.needUpdate = nil
         end
      else
         local b = famBags.Buttons:CreateEmpty(i, bag, bagType)
         b.info.needUpdate = nil
      end
      if not famBags.Buttons[i] then
         famBags.Buttons:CreateBag(i)
      end
      famBags.Buttons[i].bagType = bagType
   end
   famBags:LayoutEmptiesButtons(bag)
   self.EMPTIES_NEED_LAYOUT[bag] = nil
end

function famBags:LayoutEmptiesButtons(bag)
   if not (famBags.Buttons.empties and famBags.Buttons.empties[bag]) then return end
   local b1, count
   famBags.Sections[bag].empties:SetHeight((rawget(famBags.Buttons.empties[bag], next(famBags.Buttons.empties[bag])):GetHeight() + (db.layout.buttons.vpadding * 2)) * db.layout.buttons.size)
   if db.layout.general.sections ~= 'bags' then
      b1 = famBags:LayoutEmptiesButton(bag, famBags.Buttons.empties[bag][0], nil, 0, true)
      for n, b in pairs(famBags.Buttons.empties[bag]) do
         if n ~= 0 then
            if b.info.needUpdate then
               b = b:GetParent():Hide() -- dummyparent
               famBags.Buttons.empties[bag][n] = nil
            else
               b1 = famBags:LayoutEmptiesButton(bag, b, b1, n, nil)
            end
         end
      end
   else
      for n, b in pairs(famBags.Buttons.empties[bag]) do
         if b.info.needUpdate then
            b = b:GetParent():Hide() -- dummyparent
            famBags.Buttons.empties[bag][n] = nil
         end
      end
      local b
      local first = true
      for _, i in ipairs(self.BagIndexes[bag]) do
         b = famBags.Buttons.empties[bag][i]
         b1 = famBags:LayoutEmptiesButton(bag, b, b1, i, first)
         first = nil
      end
   end
end

function famBags:LayoutEmptiesButton(bag, b, b1, i, first)
   local count
   local db = db.layout
   if first then
      count = _G[b:GetName().."Count"]
      self:SetTextLayout(b, count, db.counts, nil, false, false)
      b = b:GetParent()
      b:ClearAllPoints()
      b:SetPoint('TOP'..db.empties.hpos, self.Sections[bag].empties, 'TOP'..db.empties.hpos, (db.empties.hpos == 'LEFT') and db.buttons.hpadding or -db.buttons.hpadding, -db.buttons.vpadding)
   else
      count = _G[b:GetName().."Count"]
      self:SetTextLayout(b, count, db.counts, nil, false, false)
      b = b:GetParent() -- dummyparent
      b:ClearAllPoints()
      b:SetPoint('TOP'..db.empties.hpos, b1, 'TOP'..(db.empties.hpos == 'LEFT' and 'RIGHT' or 'LEFT'), (db.empties.hpos == 'LEFT') and db.buttons.hspacing or -db.buttons.hspacing, 0)
   end
   return b
end

function famBags:UpdateAllButtons()
   for _, i in ipairs(self.BagIndexes['bags']) do
      self:UpdateBagButtons(i)
   end
   local slots
   for _, i in ipairs(self.BagIndexes['bank']) do
      if self.Buttons[i] then
         self:UpdateBagButtons(i, #self.Buttons[i])
      end
   end
   self.CHANGING_PROFILE = nil
end

function famBags:LightLayoutBagsButtons(bag)
   self:UpdateEmptiesButtons(bag)
   if db.layout.general.sections == 'categories' then
      famBags.Buttons:UpdateCurrentFromCategories(bag)
   else
      famBags.Buttons:UpdateCurrentFromBags(bag, db.layout.general.sections == 'onebag')
   end
end

function famBags:LayoutBagsButtons(bag)
   if db.layout.general.sections == 'onebag' then
      if bag then
         self:LayoutButtonsAsOneBag(bag)
      else
         self:LayoutButtonsAsOneBag('bags')
         self:LayoutButtonsAsOneBag('bank')
      end
   else
      if bag then
         self:LayoutButtonsBySections(bag)
      else
         self:LayoutButtonsBySections('bags')
         self:LayoutButtonsBySections('bank')
      end
   end
end

function famBags:LayoutButtonsBySections(bag)
   if bag == 'bank' and (not self.famBankFrame:IsVisible()) then
      return
   end
   self:UpdateEmptiesButtons(bag)
   local max_x_row = db[(bag == 'bags') and 'mainwindow' or 'bank'].layout.buttons_x_row
   local hp = db.layout.buttons.hpadding
   local vp = db.layout.buttons.vpadding
   local hs = db.layout.buttons.hspacing
   local vs = db.layout.buttons.vspacing
   local size = db.layout.buttons.size
   local ordered = db.layout.general.ordered
   local sectionsType = db.layout.general.sections
   
   famBags.Buttons:ClearAllPoints(bag)
   if sectionsType == 'categories' then
      self.Buttons:BuildCurrentFromCategories(bag, ordered)
   else
      self.Buttons:BuildCurrentFromBags(bag, ordered, false)
   end
   
   local bw = nil
   local bh = nil
   for i, buttons in pairs(self.Buttons.Current[bag]) do
      local slots = #buttons
      if slots and slots > 0 and self.Sections[bag][sectionsType][i] then
         local section = self.Sections[bag][sectionsType][i]
         if sectionsType == 'bags' then
            self.Sections:SetTitle(sectionsType, section, GetBagName(section.name))
         end
      
         local rowpos = 1
         local maxrowpos = 1
         local rows = 1
         local th = section.titleFrame:GetHeight()
         local ti = vp
         local dbsection = db.sections[(sectionsType == 'bags') and 'bags' or i]
         local ox
         local oy
         if dbsection.title.is_inset then
            th = th + dbsection.title.inset + vp
            ti = th
         end
         for j=1, slots do
            if rowpos > maxrowpos then
               maxrowpos = rowpos
            end
            local b = buttons[j]
            if (not bw) then
               bw = b:GetWidth()
               bh = b:GetHeight()
            end
            ox = hp + ((rowpos-1) * (hs + bw))
            oy = -(ti + ((rows-1) * (vs + bh)))
            b:SetPoint('TOPLEFT', section.bodyFrame, 'TOPLEFT', ox, oy)
            for _, b1 in ipairs(b.stacks) do
               b:SetPoint('TOPLEFT', section.bodyFrame, 'TOPLEFT', ox, oy)
            end
            if rowpos >= max_x_row and j > 1 and j < slots then
               rows = rows + 1
               rowpos = 0
            end
            b:Show()
            rowpos = rowpos + 1
         end
         local tw = section.titleFrame.titleText:GetWidth()
         local w = ((bw * maxrowpos) + (hs * (maxrowpos - 1)) + (hp * 2)) * size
         section:SetWidth(w)
         if w < tw + (dbsection.title.text.padding * 2) then
            local h = section.titleFrame.titleText:GetHeight()
            section.titleFrame.titleText:SetWidth(w - (dbsection.title.text.padding * 2))
            section.titleFrame.titleText:SetHeight(h)
         end
         section:SetHeight(th + (((bh * rows) + (vs * (rows - 1)) + (vp * 2)) * size))
         section:Show()
      elseif self.Sections[bag][sectionsType][i] then
         self.Sections[bag][sectionsType][i]:Hide()
      end
   end
   if not self.isStackMode then
      self.Sections[bag].empties:Hide()
   end
   famBags.Sections:LayoutBags(bag, sectionsType)
end

function famBags:LayoutButtonsAsOneBag(bag)
   if bag == 'bank' and (not self.famBankFrame:IsVisible()) then
      return
   end
   self:UpdateEmptiesButtons(bag)
   local max_x_row = db[(bag == 'bags') and 'mainwindow' or 'bank'].layout.buttons_x_row
   local hp = db.layout.buttons.hpadding
   local vp = db.layout.buttons.vpadding
   local hs = db.layout.buttons.hspacing
   local vs = db.layout.buttons.vspacing
   local size = db.layout.buttons.size
   local ordered = db.layout.general.ordered

   famBags.Buttons:ClearAllPoints(bag)
   for _, i in ipairs(famBags.BagIndexes[bag]) do
      if self.Sections[bag]['bags'] and self.Sections[bag]['bags'][i] then
         self.Sections[bag]['bags'][i]:Hide()
      end
   end

   self.Buttons:BuildCurrentFromBags(bag, ordered, true)
   local buttons = self.Buttons.Current[bag]

   local rowpos = 1
   local maxrowpos = 1
   local rows = 1
   local bw = 0
   local bh = 0
   local dbbag = (bag == 'bags') and 'mainwindow' or 'bank'
   local f = (bag == 'bags') and famBags.famBagsFrame or famBags.famBankFrame
   local th = f.titleFrame:GetHeight()
   local ti = 0
   if db[dbbag].title.is_inset then
      th = th + (db[dbbag].title.inset)
      ti = th
   end
   if famBags.isStackMode then
      local dbe = db.layout.empties
      local s = famBags.Sections[bag].empties
      s:ClearAllPoints()
      s:SetPoint(dbe.vpos..'LEFT', f.bodyFrame, dbe.vpos..'LEFT', dbe.hpadding, (dbe.vpos == 'TOP') and -(ti + dbe.vpadding) or dbe.vpadding)
      s:SetPoint('RIGHT', f.bodyFrame, 'RIGHT', -dbe.hpadding, 0)
      if dbe.vpos == 'TOP' then
         ti = ti + s:GetHeight()/size + dbe.vpadding
      end
      th = th + s:GetHeight() + dbe.vpadding
      s:Show()
   end
   for i=1, #buttons do
      maxrowpos = max(maxrowpos, rowpos)
      local b = buttons[i]
      if i == 1 then
         b:SetPoint('TOPLEFT', f.bodyFrame, 'TOPLEFT', hp, -(vp + ti))
         bw = b:GetWidth()
         bh = b:GetHeight()
      elseif rowpos == 1 then
         b:SetPoint('TOPLEFT', buttons[i - max_x_row], 'BOTTOMLEFT', 0, -vs)
         rows = rows + 1
      else
         b:SetPoint('TOPLEFT', buttons[i-1], 'TOPRIGHT', hs, 0)
         if rowpos == max_x_row then
            rowpos = 0
         end
      end
      b:Show()
      rowpos = rowpos + 1
   end
   local dbw = famBags.db.profile[dbbag].layout
   if dbw.fixed_width then
      f:SetWidth(((bw * dbw.buttons_x_row) + (hs * (dbw.buttons_x_row - 1)) + (hp * 2)) * size)
   else
      f:SetWidth(((bw * maxrowpos) + (hs * (maxrowpos - 1)) + (hp * 2)) * size)
   end
   f:SetHeight((th + ((bh * rows) + (vs * (rows - 1)) + (vp * 2))) * size)
end

function famBags:CreateBagFrame(name, title, parent, startLevel, w, h, db, movable)
   local f = (_G[name]) and _G[name] or CreateFrame('Frame', name, parent)
   
   if movable then
      f:SetMovable(true)
      f:EnableMouse(true)
      f:RegisterForDrag('LeftButton')
      f:SetScript('OnMouseDown', function(f)
         f.pOrigin = f.pOrigin or {}
         f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, f.pOrigin.x, f.pOrigin.y = f:GetPoint(1)
         f:StartMoving()
         f.pStart = f.pStart or {}
         f.pStart.x, f.pStart.y = select(4, f:GetPoint(1))
      end)
      f:SetScript('OnMouseUp', function(f)
         f.pEnd = f.pEnd or {}
         f.pEnd.x, f.pEnd.y = select(4, f:GetPoint(1))
         f:StopMovingOrSizing()
         f:ClearAllPoints()
         f:SetPoint(f.pOrigin.point, f.pOrigin.anchor, f.pOrigin.relPoint, f.pOrigin.x + (f.pEnd.x - f.pStart.x), f.pOrigin.y + (f.pEnd.y - f.pStart.y))
      end)
   end
   
   f:SetFrameStrata(self.db.profile.layout.general.strata)

   if f:GetEffectiveScale() == 1 then
      f:SetScale(UIParent:GetEffectiveScale())
   end
   f:SetWidth(w)
   f:SetHeight(h)

   f.bodyFrame = f.bodyFrame or CreateFrame('Frame', name..'_Body', f)
   if f.bodyFrame:GetEffectiveScale() == 1 then
      f.bodyFrame:SetScale(UIParent:GetEffectiveScale())
   end
   f.bodyFrame:SetFrameLevel(startLevel)
   self:SetFrameBackdrop(f.bodyFrame, db)

   if db.title then
      f.titleFrame = f.titleFrame or CreateFrame('Frame', name..'_Title', f)
      if f.titleFrame:GetEffectiveScale() == 1 then
         f.titleFrame:SetScale(UIParent:GetEffectiveScale())
      end
      f.titleFrame:SetFrameLevel(startLevel + 1)
      self:SetFrameBackdrop(f.titleFrame, db.title)
      self:SetTextLayout(f.titleFrame, 'titleText', db.title.text, title, true, false)
      self:SetTitleInset(f, db.title.is_inset, db.title.inset, db.title.text.padding)
   else
      f.bodyFrame:SetAllPoints(f)
   end

   return f
end

function famBags:SetTextLayout(frame, fontstring, db, text, autoheightparent, autowidthparent)
   local t
   if type(fontstring) == 'string' then
      if not frame[fontstring] then
         frame[fontstring] = frame:CreateFontString(nil, 'OVERLAY');
         frame[fontstring].autoheight = autoheightparent
         frame[fontstring].autowidth = autowidthparent
      end
      t = frame[fontstring]
   else
      t = fontstring
   end

   local flags = ''
   for n, v in pairs(db.flags) do
      if n ~= 'SHADOW' and v then
         if flags ~= '' then
            flags = flags..', '
         end
         flags = flags..n
      end
   end
   t:SetFont(LSM:Fetch('font', db.font or '') or DEFAULT_FONT, db.size, flags)
   if db.flags.SHADOW then
      t:SetShadowColor(0, 0, 0, 1)
      t:SetShadowOffset(0.8, -0.8)
   elseif t:GetShadowOffset() then
      t:SetShadowColor(0, 0, 0, 0)
      t:SetShadowOffset(0, 0)
   end

   t:SetWidth(0)
   t:SetHeight(0)
   t:ClearAllPoints()
   local align, h, v = get_point(db.halign, db.valign, db.padding)
   t:SetPoint(align, frame, align, h, v);
   if text ~= nil then
      t.text = text
   end
   if t.text then
      t:SetText(string.format('|cff%02x%02x%02x%s|r', db.color[1]*255, db.color[2]*255, db.color[3]*255, t.text));
   end

   if t.autoheight then
      frame:SetHeight(t:GetHeight() + db.padding*2 + 1)
   end

   if t.autowidth then
      frame:SetWidth(t:GetWidth() + db.padding*2)
   elseif frame:GetWidth() > 0 and frame:GetWidth() < t:GetWidth() + db.padding*2 then
      t:SetWidth(frame:GetWidth() - db.padding*2)
      t:SetHeight(frame:GetHeight() - db.padding*2)
   end
end

function famBags:SetTitleInset(frame, is_inset, inset, padding)
   frame.titleFrame:ClearAllPoints()
   frame.bodyFrame:ClearAllPoints()
   if is_inset then
      frame.titleFrame:SetPoint('TOPLEFT', frame, 'TOPLEFT', inset, -inset)
      frame.titleFrame:SetPoint('RIGHT', frame, 'RIGHT', -inset, 0)
      frame.titleFrame:SetHeight(frame.titleFrame.titleText:GetHeight() + (padding*2) + 1)
      frame.bodyFrame:SetAllPoints(frame)
   else
      frame.titleFrame:SetPoint('TOPLEFT', frame, 'TOPLEFT', 0, 0)
      frame.titleFrame:SetHeight(frame.titleFrame.titleText:GetHeight() + (padding*2) + 1)
      if frame.titleFrame.titleText.autowidth then
         frame.titleFrame:SetWidth(frame.titleFrame.titleText:GetWidth() + (padding*2))
      else
         frame.titleFrame:SetPoint('RIGHT', frame, 'RIGHT', 0, 0)
      end
      frame.bodyFrame:SetPoint('TOPLEFT', frame.titleFrame, 'BOTTOMLEFT', 0, 0)
      frame.bodyFrame:SetPoint('BOTTOMRIGHT', frame, 'BOTTOMRIGHT', 0, 0)
   end
end

function famBags:SetFrameBackdrop(frame, db)
   if not frame.bg then
      frame.bg = frame:CreateTexture(nil, 'BACKGROUND')
   end
   local t = frame.bg
   t:SetTexture(1,1,1,1)
   if db.bg.type == 1 --[[ Solid--]] then
      t:SetGradientAlpha('VERTICAL',
         db.bg.color1[1], db.bg.color1[2], db.bg.color1[3], db.bg.color1[4],
         db.bg.color1[1], db.bg.color1[2], db.bg.color1[3], db.bg.color1[4])
   elseif db.bg.type == 2 --[[ Vertical gradient--]] then
      t:SetGradientAlpha('VERTICAL',
         db.bg.color1[1], db.bg.color1[2], db.bg.color1[3], db.bg.color1[4],
         db.bg.color2[1], db.bg.color2[2], db.bg.color2[3], db.bg.color2[4])
   elseif db.bg.type == 3 --[[ Horizontal gradient--]] then
      t:SetGradientAlpha('HORIZONTAL',
         db.bg.color1[1], db.bg.color1[2], db.bg.color1[3], db.bg.color1[4],
         db.bg.color2[1], db.bg.color2[2], db.bg.color2[3], db.bg.color2[4])
   elseif db.bg.type == 4 --[[ Texture--]] then
      t:SetTexture(LSM:Fetch('background', db.bg.texture))
      t:SetVertexColor(unpack(db.bg.color1))
   end
   t:SetPoint('TOPLEFT', frame, 'TOPLEFT', db.bg.inset, -db.bg.inset)
   t:SetPoint('BOTTOMRIGHT', frame, 'BOTTOMRIGHT', -db.bg.inset, db.bg.inset)
   -- SetBackdrop only for border
   frame:SetBackdrop(
      {edgeFile = LSM:Fetch('border', db.bg.br_texture),
       edgeSize = db.bg.br_size});
   frame:SetBackdropBorderColor(unpack(db.bg.br_color));
end

function famBags:GetMoney()
   local copper = GetMoney()
   local size = db.money.text.size + 1
   local t = string.format('%d|TInterface\\MoneyFrame\\UI-GoldIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper / 100 / 100, size, size)
   if db.money.show_silver then
      t = t..string.format(' %d|TInterface\\MoneyFrame\\UI-SilverIcon:%d:%d:3:0:64:64:4:60:4:60|t ', (copper / 100) % 100, size, size)
   end
   if db.money.show_copper then
      t = t..string.format(' %d|TInterface\\MoneyFrame\\UI-CopperIcon:%d:%d:3:0:64:64:4:60:4:60|t ', copper % 100, size, size)
   end

   for i=1, GetCurrencyListSize() do
      local name, isHeader, _, _, isWatched, count, icon = GetCurrencyListInfo(i)
      if (not isHeader) and isWatched then
         t = (count or 0)..'|T'..icon..':'..size..':'..size..':3:0:64:64:4:60:4:60|t   '..t
      end
   end
   
   return t
end

function famBags:CreateMoneyFrame(db)
   local f = _G['famBagsMoneyFrame'] or CreateFrame('Frame', 'famBagsMoneyFrame', self.famBagsFrame)
   if f:GetEffectiveScale() == 1 then
      f:SetScale(UIParent:GetEffectiveScale())
   end
   f:SetPoint('TOPLEFT', self.famBagsFrame, 'BOTTOMLEFT', 0, 0)
   f:SetPoint('RIGHT', self.famBagsFrame, 'RIGHT', 0, 0)
   f:SetHeight(10)
   self:SetFrameBackdrop(f, db)

   self:SetTextLayout(f, 'moneyText', db.text, self:GetMoney(), true, false)

   return f
end

function famBags:CreateBlizzardBankFrame(db)
   local f
   if _G['famBagsBlizzardBankFrame'] then
      f = _G['famBagsBlizzardBankFrame']
      self:SetFrameBackdrop(f, db)
      return f
   end

   f = CreateFrame('Frame', 'famBagsBlizzardBankFrame', self.famBankFrame)
   if f:GetEffectiveScale() == 1 then
      f:SetScale(UIParent:GetEffectiveScale())
   end
   f:SetPoint('TOPLEFT', self.famBankFrame, 'BOTTOMLEFT', 0, 0)
   f:SetPoint('RIGHT', self.famBankFrame, 'RIGHT', 0, 0)
   self:SetFrameBackdrop(f, db)
   
   f.BankBags = {}
   for i=1, 7 do
      local b = 'BankFrameBag'..i
      f.BankBags[i] = _G[b]
      f.BankBags[i]:SetParent(f)
      f.BankBags[i]:ClearAllPoints()
      if lbf and famBags.BFBags then
         famBags.BFBags:AddButton(f.BankBags[i],{
            Icon = _G[b.."IconTexture"]
         })
      end
   end
   f:SetHeight(f.BankBags[1]:GetHeight() + 10)
   f.BankBags[4]:SetPoint('CENTER', f, 'CENTER', 0, 0)
   f.BankBags[4]:SetPoint('TOP', f, 'TOP', 0, -10)
   f.BankBags[3]:SetPoint('TOPRIGHT', f.BankBags[4], 'TOPLEFT', -5, 0)
   f.BankBags[2]:SetPoint('TOPRIGHT', f.BankBags[3], 'TOPLEFT', -5, 0)
   f.BankBags[1]:SetPoint('TOPRIGHT', f.BankBags[2], 'TOPLEFT', -5, 0)
   f.BankBags[5]:SetPoint('TOPLEFT', f.BankBags[4], 'TOPRIGHT', 5, 0)
   f.BankBags[6]:SetPoint('TOPLEFT', f.BankBags[5], 'TOPRIGHT', 5, 0)
   f.BankBags[7]:SetPoint('TOPLEFT', f.BankBags[6], 'TOPRIGHT', 5, 0)


   f.PurchaseInfo = _G['BankFramePurchaseInfo']
   local _, full = GetNumBankSlots()
   if not full then
      f.PurchaseInfo:SetParent(f)
      f.PurchaseInfo:SetPoint('CENTER', f, 'CENTER', -10, 0)
      f.PurchaseInfo:SetPoint('BOTTOM', f, 'BOTTOM', 0, -5)
      f:SetHeight(f:GetHeight() + f.PurchaseInfo:GetHeight())
   else
      f:SetHeight(f:GetHeight() + 10)
   end

   return f
end

function famBags:CreateBagSlotsFrame(db)
   local f
   if _G['famBagsBagSlotsFrame'] then
      f = _G['famBagsBagSlotsFrame']
      self:SetFrameBackdrop(f, db)
      return f
   end

   f = CreateFrame('Frame', 'famBagsBagSlotsFrame', self.famBagsFrame)
   if f:GetEffectiveScale() == 1 then
      f:SetScale(UIParent:GetEffectiveScale())
   end
   f:SetPoint('TOPRIGHT', self.famBagsFrame, 'TOPLEFT', 0, 0)
   self:SetFrameBackdrop(f, db)
   
   f.BagSlots = {}
   
   local function hack_slot(i, name)
      f.BagSlots[i] = _G[name]
      f.BagSlots[i]:SetParent(f)
      f.BagSlots[i]:ClearAllPoints()
      if lbf and famBags.BFBags then
         if i ~= KEYRING_CONTAINER then
            famBags.BFBags:AddButton(f.BagSlots[i],{
               Icon = _G[name.."IconTexture"]
            })
         else
            f.KeyRingTexture = f.BagSlots[i]:GetNormalTexture()
            famBags.BFBags:AddButton(f.BagSlots[i],{
               Icon = f.KeyRingTexture
            })
         end
      end
   end
   
   for i=0, NUM_BAG_SLOTS do
      local name = (i > 0) and 'CharacterBag'..(i-1)..'Slot' or 'MainMenuBarBackpackButton'
      hack_slot(i, name)
      if i > 0 then
         f.BagSlots[i]:SetPoint('TOPLEFT', f.BagSlots[i-1], 'BOTTOMLEFT', 0, (i > 1) and -3 or -4)
      else
         f.BagSlots[i]:SetPoint('TOPLEFT', f, 'TOPLEFT', 6, -6)
      end
   end
   
   hack_slot(KEYRING_CONTAINER, 'KeyRingButton')
   f.BagSlots[KEYRING_CONTAINER]:SetPoint('TOPLEFT', f.BagSlots[1], 'TOPRIGHT', 2, 0)
   local bs = f.BagSlots[1]:GetWidth() + f.BagSlots[KEYRING_CONTAINER]:GetWidth() + 2
   f.BagSlots[0]:SetScale(bs / f.BagSlots[0]:GetWidth())
   
   f:SetWidth(bs + 15)
   f:SetHeight(bs + ((f.BagSlots[1]:GetHeight() + 3) * NUM_BAG_SLOTS) + 16)

   return f
end

function famBags:CreateCheckButton(name, frame, tnormal, tpushed, tchecked, thighlight, tooltip, onclick)
   local b = CreateFrame('CheckButton', name, frame, 'ChatConfigCheckButtonTemplate')
   b:SetNormalTexture(tnormal)
   b:SetPushedTexture(tpushed)
   b:SetCheckedTexture(tchecked)
   b:SetHighlightTexture(thighlight)
   b:SetHitRectInsets(4,4,4,4)
   b.tooltip = tooltip
   b:SetScript("OnClick", onclick)
   
   return b
end

function famBags:CreateNormalButton(name, frame, onclick, tnormal, tpushed, thighlight, tooltip)
   local b = CreateFrame('Button', name, frame, 'UIPanelCloseButton')
   b:SetHitRectInsets(4,4,4,4)
   b:SetScript("OnClick", onclick)
   if tnormal then
      b:SetNormalTexture(tnormal)
      b:SetPushedTexture(tpushed)
      b:SetHighlightTexture(thighlight)
      b.tooltip = tooltip
   end
   b:SetScript("OnEnter", 
      function(self) 
         if (self.tooltip) then 
            GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
            GameTooltip:SetText(self.tooltip, nil, nil, nil, nil, (self.tooltipStyle or 1))
         end
      end)
   b:SetScript("OnLeave", 
      function(self) 
         GameTooltip:Hide()
      end)
   
   return b
end

function famBags:CreateIntefaceButtons()
   self.BagsCloseButton = self.BagsCloseButton or self:CreateNormalButton(
      'famBagsBagsCloseButton', 
      famBags.famBagsFrame, 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags:HideBags()
      end)
   self.BagsCloseButton:SetPoint('TOPRIGHT', famBags.famBagsFrame.titleFrame, 'TOPRIGHT', 4, 4)
   self.BagsCloseButton:SetPoint('BOTTOM', famBags.famBagsFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BagsCloseButton:SetPoint('LEFT', famBags.famBagsFrame.titleFrame, 'RIGHT', -famBags.famBagsFrame.titleFrame:GetHeight() - 2, 0)
   self.BagsCloseButton:Show()

   self.BagsStackButton = self.BagsStackButton or self:CreateCheckButton(
      'famBagsBagsStackButton', 
      famBags.famBagsFrame, 
      'Interface\\Addons\\famBags\\Textures\\btnnostack.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnnostack.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnstack.blp', 
      'Interface/Buttons/UI-Panel-MinimizeButton-Highlight', 
      L['Stacking mode on/off'], 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags.isStackMode = self:GetChecked()
         famBags.BankStackButton:SetChecked(self:GetChecked())
         famBags:LayoutBagsButtons()
      end)
   self.BagsStackButton:SetPoint('TOP', famBags.famBagsFrame.titleFrame, 'TOP', 0, 4)
   self.BagsStackButton:SetPoint('BOTTOM', famBags.famBagsFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BagsStackButton:SetPoint('RIGHT', self.BagsCloseButton, 'LEFT', 5, 0)
   self.BagsStackButton:SetPoint('LEFT', self.BagsCloseButton, 'LEFT', -famBags.famBagsFrame.titleFrame:GetHeight() - 2, 0)
   self.BagsStackButton:Show()
   
   self.BagsRefreshButton = self.BagsRefreshButton or self:CreateNormalButton(
      'famBagsBagsRefreshButton', 
      famBags.famBagsFrame, 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags:LayoutBagsButtons('bags')
      end,
      'Interface\\Addons\\famBags\\Textures\\btnrefresh.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnrefreshdown.blp', 
      'Interface/Buttons/UI-Panel-MinimizeButton-Highlight',
      L['Refresh Layout'])
   self.BagsRefreshButton:SetPoint('TOP', famBags.famBagsFrame.titleFrame, 'TOP', 0, 4)
   self.BagsRefreshButton:SetPoint('BOTTOM', famBags.famBagsFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BagsRefreshButton:SetPoint('RIGHT', self.BagsStackButton, 'LEFT', 5, 0)
   self.BagsRefreshButton:SetPoint('LEFT', self.BagsStackButton, 'LEFT', -famBags.famBagsFrame.titleFrame:GetHeight() - 2, 0)
   self.BagsRefreshButton:Show()


   if db.bagswindow.enabled then
      self.BagSlotsButton = self.BagSlotsButton or self:CreateCheckButton(
         'famBagsBagSlotsButton', 
         famBags.famBagsFrame, 
         'Interface\\Addons\\famBags\\Textures\\btnshowbags.blp', 
         'Interface\\Addons\\famBags\\Textures\\btnshowbags.blp', 
         'Interface\\Addons\\famBags\\Textures\\btnhidebags.blp', 
         'Interface/Buttons/UI-Panel-MinimizeButton-Highlight', 
         L['Bag slots on/off'], 
         function(self)
            local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
            if self:GetChecked() then
               famBags.famBagsBagSlotsFrame:Show()
            else
               famBags.famBagsBagSlotsFrame:Hide()
            end
         end)
      self.BagSlotsButton:SetPoint('TOP', famBags.famBagsFrame.titleFrame, 'TOP', 0, 4)
      self.BagSlotsButton:SetPoint('BOTTOM', famBags.famBagsFrame.titleFrame, 'BOTTOM', 0, -2)
      self.BagSlotsButton:SetPoint('RIGHT', self.BagsRefreshButton, 'LEFT', 5, 0)
      self.BagSlotsButton:SetPoint('LEFT', self.BagsRefreshButton, 'LEFT', -famBags.famBagsFrame.titleFrame:GetHeight() - 2, 0)
      self.BagSlotsButton:Show()
   end

   self.BagsResetUpdatedButton = self.BagsResetUpdatedButton or self:CreateNormalButton(
      'famBagsBagsResetUpdatedButton', 
      famBags.famBagsFrame, 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags.Buttons:ResetUpdatedIndicators()
         famBags:LayoutBagsButtons('bags')
      end,
      'Interface\\Addons\\famBags\\Textures\\btnreset.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnresetdown.blp', 
      'Interface/Buttons/UI-Panel-MinimizeButton-Highlight',
      L['Reset updated item indicators'])
   
   local anchorframe = db.bagswindow.enabled and self.BagSlotsButton or self.BagsStackButton
   
   self.BagsResetUpdatedButton:SetPoint('TOP', famBags.famBagsFrame.titleFrame, 'TOP', 0, 4)
   self.BagsResetUpdatedButton:SetPoint('BOTTOM', famBags.famBagsFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BagsResetUpdatedButton:SetPoint('RIGHT', anchorframe, 'LEFT', 5, 0)
   self.BagsResetUpdatedButton:SetPoint('LEFT', anchorframe, 'LEFT', -famBags.famBagsFrame.titleFrame:GetHeight() - 2, 0)
   self.BagsResetUpdatedButton:Show()


   self.BankCloseButton = self.BankCloseButton or self:CreateNormalButton(
      'famBagsBankCloseButton', 
      famBags.famBankFrame, 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         CloseBankFrame()
      end)
   self.BankCloseButton:SetPoint('TOPRIGHT', famBags.famBankFrame.titleFrame, 'TOPRIGHT', 4, 4)
   self.BankCloseButton:SetPoint('BOTTOM', famBags.famBankFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BankCloseButton:SetPoint('LEFT', famBags.famBankFrame.titleFrame, 'RIGHT', -famBags.famBankFrame.titleFrame:GetHeight() - 2, 0)
   self.BankCloseButton:Show()

   self.BankStackButton = self.BankStackButton or self:CreateCheckButton(
      'famBagsBankStackButton', 
      famBags.famBankFrame, 
      'Interface\\Addons\\famBags\\Textures\\btnnostack.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnnostack.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnstack.blp', 
      'Interface/Buttons/UI-Panel-MinimizeButton-Highlight', 
      L['Stacking mode on/off'], 
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags.isStackMode = self:GetChecked()
         famBags.BagsStackButton:SetChecked(self:GetChecked())
         famBags:LayoutBagsButtons()
      end)
   self.BankStackButton:SetPoint('TOP', famBags.famBankFrame.titleFrame, 'TOP', 0, 4)
   self.BankStackButton:SetPoint('BOTTOM', famBags.famBankFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BankStackButton:SetPoint('RIGHT', self.BankCloseButton, 'LEFT', 5, 0)
   self.BankStackButton:SetPoint('LEFT', self.BankCloseButton, 'LEFT', -famBags.famBankFrame.titleFrame:GetHeight() - 2, 0)
   self.BankStackButton:Show()

   self.BankRefreshButton = self.BankRefreshButton or self:CreateNormalButton(
      'famBagsBankRefreshButton', 
      famBags.famBankFrame,            
      function(self)
         local famBags = LibStub('AceAddon-3.0'):GetAddon('famBags')
         famBags:LayoutBagsButtons('bank')
      end,
      'Interface\\Addons\\famBags\\Textures\\btnrefresh.blp', 
      'Interface\\Addons\\famBags\\Textures\\btnrefreshdown.blp', 
      'Interface/Buttons/UI-Panel-MinimizeButton-Highlight',
      L['Refresh Layout'])
   self.BankRefreshButton:SetPoint('TOP', famBags.famBankFrame.titleFrame, 'TOP', 0, 4)
   self.BankRefreshButton:SetPoint('BOTTOM', famBags.famBankFrame.titleFrame, 'BOTTOM', 0, -2)
   self.BankRefreshButton:SetPoint('RIGHT', self.BankStackButton, 'LEFT', 5, 0)
   self.BankRefreshButton:SetPoint('LEFT', self.BankStackButton, 'LEFT', -famBags.famBankFrame.titleFrame:GetHeight() - 2, 0)
   self.BankRefreshButton:Show()

end


--[[
function famBags:OnDisable()
end

SLASH_FAMBAGS1 = '/fambags'
SlashCmdList['FAMBAGS'] = function(msg, editbox)
end
--]]